@@ -157,10 +157,12 @@ class MSP430Assembler(object):
 
     def argument(self, value):
         """\
-        Prepare a parameter for the linker. Numbers (nnn, 0xnnn) are supported
-        as well as labels.
+        Prepare a parameter for the linker. Expressions are evaluated.
         """
-        return infix2postfix(value, variable_prefix=u'GET-SYMBOL ')
+        try:
+            return infix2postfix(value, variable_prefix=u'GET-SYMBOL ')
+        except ValueError, e:
+            raise AssemblerError('invalid expression: %r' % (value,))
 
     # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
@@ -204,25 +206,27 @@ class MSP430Assembler(object):
                         return (3, 3, None, 0)
             return (3, 0, match_obj.group('IMM_VAL'), 0)
 
-        if mode == 'ABSOLUTE':
-            return (1, 2, match_obj.group('ABS_VAL'), 0)
+        try:
+            if mode == 'ABSOLUTE':
+                return (1, 2, match_obj.group('ABS_VAL'), 0)
 
-        if mode == 'INDEXED':
-            return (1, regnumbers[match_obj.group('IDX_REG')], match_obj.group('IDX_VALUE'), 0)
+            if mode == 'INDEXED':
+                return (1, regnumbers[match_obj.group('IDX_REG')], match_obj.group('IDX_VALUE'), 0)
 
-        if mode == 'POST_INC':
-            return (3, regnumbers[match_obj.group('PI_REG')], None, 0)
+            if mode == 'POST_INC':
+                return (3, regnumbers[match_obj.group('PI_REG')], None, 0)
 
-        if mode == 'INDIRECT':
-            return (2, regnumbers[match_obj.group('IND_REG')], None, 0)
+            if mode == 'INDIRECT':
+                return (2, regnumbers[match_obj.group('IND_REG')], None, 0)
 
-        if mode == 'REGISTER':
-            if value.upper() in regnumbers:         # register mode
-                return (0, regnumbers[value.upper()], None, 0)
-
-        if mode == 'SYMBOLIC':
-            return (1, 0, value, 1)         # symbolic mode
+            if mode == 'REGISTER':
+                if value.upper() in regnumbers:         # register mode
+                    return (0, regnumbers[value.upper()], None, 0)
 
+            if mode == 'SYMBOLIC':
+                return (1, 0, value, 1)         # symbolic mode
+        except KeyError as e:
+            raise AssemblerError('Register name invalid: %s' % e)
         raise AssemblerError('Bad argument type: %s %s' % (mode, value))
 
 
@@ -279,7 +283,7 @@ class MSP430Assembler(object):
         asrc, src, op1, rel1 = self._buildArg(insn, source)
         adst, dst, op2, rel2 = self._buildArg(insn, destination, constreg=False)
         if adst > 1: raise AssemblerError('%s (%r) is not possible as destination.' % (
-                self._name_address_mode(adst,dst), destination))
+                self._name_address_mode(adst,dst), destination[1]))
         out = [u'0x%04x 16bit' % self._buildDoubleOperand(insn, bw, asrc, src, adst, dst)]
         if op1:
             if rel1: out.append(u'%s PC - 16bit' % self.argument(op1))
@@ -297,7 +301,7 @@ class MSP430Assembler(object):
         asrc, src, op1, rel1 = self._buildArg(insn, source)
         adst, dst, op2, rel2 = self._buildArg(insn, destination, constreg=False)
         if adst > 1: raise AssemblerError('%s (%r) is not possible as destination.' % (
-                self._name_address_mode(adst,dst), destination))
+                self._name_address_mode(adst,dst), destination[1]))
         out = []
         # the core instruction
         opcode = self._buildDoubleOperand(insn[:-1], bw, asrc, src, adst, dst)
@@ -477,7 +481,7 @@ class MSP430Assembler(object):
 
         if adst != 0:
             raise AssemblerError(u'%s (%r) is not possible as destination.' % (
-                    self._name_address_mode(adst, dst), destination))
+                    self._name_address_mode(adst, dst), destination[1]))
 
         out = []
         if asrc == 0:       # register mode
@@ -490,7 +494,7 @@ class MSP430Assembler(object):
             return u' '.join(out)
         else:
             raise AssemblerError(u'%s (%r) is not possible as source.' % (
-                self._name_address_mode(asrc, src), source))
+                self._name_address_mode(asrc, src), source[1]))
 
     def insnx_ADDA_2(self, insn, mode, source, destination):
         """Add 20-bit source to a 20-bit destination register"""
@@ -519,7 +523,7 @@ class MSP430Assembler(object):
 
         if adst != 0:
             raise AssemblerError('%s (%r) is not possible as destination.' % (
-                    self._name_address_mode(adst, dst), destination))
+                    self._name_address_mode(adst, dst), destination[1]))
 
         out = []
         if asrc == 3:       # immediate/post_inc mods - register mode
@@ -534,7 +538,7 @@ class MSP430Assembler(object):
                           'RRUM': 0x0340,
                         }[insn] | (option << 4) | (count<<10) | dst)
         raise AssemblerError(u'%s (%r) is not possible as count.' % (
-                self._name_address_mode(asrc, src), source))
+                self._name_address_mode(asrc, src), source[1]))
 
 
     def insnx_CALLA_1(self, insn, mode, target):
@@ -582,7 +586,7 @@ class MSP430Assembler(object):
 
         if adst != 0:
             raise AssemblerError(u'%s (%r) is not possible as destination.' % (
-                    self._name_address_mode(adst, dst), destination))
+                    self._name_address_mode(adst, dst), destination[1]))
 
         out = []
         if asrc == 3:       # immediate/post_inc mods - register mode
@@ -1005,15 +1009,15 @@ class MSP430Assembler(object):
                                         output.write('\n')
                         else:
                             raise AssemblerError(u'Syntax Error: unknown instruction %r' % (insn,))
-        except AssemblerError, e:
+        except AssemblerError as e:
             # annotate exception with location in source file
             e.line = lineno
             e.filename = filename
+            e.text = line
             raise e
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-
-if __name__ == '__main__':
+def main():
     from optparse import OptionParser
 
     parser = OptionParser()
@@ -1095,13 +1099,24 @@ if __name__ == '__main__':
                     options.input_filename,
                     output=out)
         else:
-            f = codecs.open(filename, 'r', 'utf-8')
-            assembler.assemble(f, options.input_filename, output=out)
-            f.close()
-    except AssemblerError, e:
+            try:
+                f = codecs.open(filename, 'r', 'utf-8')
+                assembler.assemble(f, options.input_filename, output=out)
+                f.close()
+            except IOError as e:
+                sys.stderr.write('as: %s: File not found\n' % (filename,))
+                sys.exit(1)
+    except AssemblerError as e:
         sys.stderr.write('%s:%s: %s\n' % (e.filename, e.line, e))
+        if options.debug:
+            if hasattr(e, 'text'):
+                sys.stderr.write('%s:%s: input line: %r\n' % (e.filename, e.line, e.text))
         sys.exit(1)
 
 
     if options.debug:
         sys.stderr.write("%s\n" % (("END %s" % filename).center(70).replace(' ', '-')))
+
+
+if __name__ == '__main__':
+    main()
