@@ -1,3 +1,10 @@
+#!/usr/bin/env python
+# -*- coding: utf-8 -*-
+#
+# Copyright (c) 2011 Chris Liechti <cliechti@gmx.net>
+# All Rights Reserved.
+# Simplified BSD License (see LICENSE.txt for full text)
+
 """\
 Minimalistic Forth like language interpreter.
 
@@ -27,7 +34,11 @@ class ForthError(rpn.RPNError):
 
 
 class SeekableIterator(object):
-    """An iterator with the additional functionality to seek back"""
+    """\
+    An iterator with the additional functionality to adjust the read pointer
+    while it is running. This is needed to implement jumps in
+    Frame/NativeFrame.
+    """
     def __init__(self, some_list):
         self.some_list = some_list
         self.position = 0
@@ -51,9 +62,11 @@ class SeekableIterator(object):
 
 class Frame(list):
     """Storage for function definitions"""
+
     def __init__(self, name):
         list.__init__(self)
         self.name = name
+        self.use_ram = False
 
     def __call__(self, stack):
         """Execute code in frame"""
@@ -72,9 +85,19 @@ class Frame(list):
     def __repr__(self):
         return '%s[%s]' % (self.__class__.__name__, self.name,)
 
+class InterruptFrame(Frame):
+    """\
+    Interrupt frames are like normal Frames in most aspects but need different
+    entry/exit code.
+    """
+    def __init__(self, name, vector):
+        Frame.__init__(self, name)
+        self.vector = vector
+
 
 class NativeFrame(list):
     """Storage for native function definitions"""
+
     def __init__(self, name):
         list.__init__(self)
         self.name = name
@@ -98,7 +121,13 @@ class NativeFrame(list):
 
 
 class Variable(object):
-    """This emulates what on a target would be an address"""
+    """This emulates what on a target would be an address."""
+    # typical variable usage: "HERE @". so the variable name would put the
+    # address of the variable on the stack. The value of HERE is then also used
+    # to write to (e.g. in the implementation of IF/ENDIF. As we don't not have
+    # linear address space but frames for each dictionary entry that start
+    # counting at zero, the value needs to remember the frame it belongs to.
+
     def __init__(self, frame, offset):
         self.frame = frame
         self.offset = offset
@@ -143,13 +172,141 @@ def immediate(function):
     return function
 
 
-class Forth(rpn.RPN):
+class ForthBitOps(object):
+    """Forth specific words for bit operations"""
+
+    @rpn.word("OR")
+    def bitor(self, stack):
+        x, y = self.pop2()
+        self.push(y | x)
+
+    @rpn.word("AND")
+    def bitand(self, stack):
+        x, y = self.pop2()
+        self.push(y & x)
+
+    @rpn.word("XOR")
+    def bitxor(self, stack):
+        x, y = self.pop2()
+        self.push(y ^ x)
+
+    @rpn.word("LSHIFT")
+    def bit_shift_left(self, stack):
+        x, y = self.pop2()
+        self.push(y << x)
+
+    @rpn.word("RSHIFT")
+    def bit_shift_right(self, stack):
+        x, y = self.pop2()
+        self.push(y >> x)
+
+    @rpn.word("INVERT")
+    def bitnot(self, stack):
+        self.push(~self.pop())
+
+    @rpn.word("2*")
+    def arithmetic_shift_left(self, stack):
+        self[-1] = self[-1]*2
+
+    @rpn.word("2/")
+    def arithmetic_shift_right(self, stack):
+        self[-1] = self[-1]/2
+
+
+class ForthMiscOps(object):
+    """more Forth specific words"""
+
+    @rpn.word("NOT")
+    def word_NOT(self, stack):
+        self.push(not self.pop())
+
+    @rpn.word("1+")
+    def plus_1(self, stack):
+        self[-1] = self[-1] + 1
+
+    @rpn.word("2+")
+    def plus_2(self, stack):
+        self[-1] = self[-1] + 2
+
+    @rpn.word("4+")
+    def plus_2(self, stack):
+        self[-1] = self[-1] + 4
+
+    @rpn.word("1-")
+    def minus_1(self, stack):
+        self[-1] = self[-1] - 1
+
+    @rpn.word("2-")
+    def minus_2(self, stack):
+        self[-1] = self[-1] - 2
+
+    @rpn.word("4-")
+    def minus_2(self, stack):
+        self[-1] = self[-1] - 4
+
+
+    @rpn.word('/MOD')
+    def word_divmod(self, stack):
+        a = stack.pop()
+        b = stack.pop()
+        d, m = divmod(a, b)
+        stack.push(d)
+        stack.push(m)
+
+    @rpn.word('=')
+    def word_equals1(self, stack):
+        """Compare two numbers for equality"""
+        a = stack.pop()
+        b = stack.pop()
+        stack.push(a == b)
+
+    @rpn.word('0=')
+    def word_is_zero(self, stack):
+        """Check if number is not zero"""
+        a = stack.pop()
+        stack.push(a == 0)
+
+    @rpn.word('0>')
+    def word_is_positive(self, stack):
+        """Check if number is positive"""
+        a = stack.pop()
+        stack.push(a > 0)
+
+    @rpn.word('?DUP')
+    def word_Qdup(self, stack):
+        """DUP top of stack but only if not zero."""
+        if stack[-1]:
+            stack.push(stack[-1])
+
+    @rpn.word('ROT')
+    def word_is_rot(self, stack):
+        """Rotate 3 items on the stack. 3rd gets 1st."""
+        a = stack.pop()
+        b = stack.pop()
+        c = stack.pop()
+        stack.push(b)
+        stack.push(a)
+        stack.push(c)
+
+    @rpn.word('-ROT')
+    def word_is_nrot(self, stack):
+        """Rotate 3 items on the stack. 1st gets 3rd."""
+        a = stack.pop()
+        b = stack.pop()
+        c = stack.pop()
+        stack.push(a)
+        stack.push(c)
+        stack.push(b)
+
+
+class Forth(rpn.RPNBase, rpn.RPNStackOps, rpn.RPNSimpleMathOps,
+            rpn.RPNCompareOps, ForthBitOps, ForthMiscOps):
     """\
     Extension of the RPN calculator with Forth like language features.
     """
     def __init__(self, namespace=None):
-        rpn.RPN.__init__(self, namespace)
-        self.target_namespace = {}  # an other namespace for target only objects
+        rpn.RPNBase.__init__(self, namespace)
+        self.target_namespace = {}  # an other name space for target only objects
         self.compiling = False
         self.output = sys.stdout
         self.frame = None
@@ -159,12 +316,49 @@ class Forth(rpn.RPN):
         self.compiled_words = set()
         self.not_yet_compiled_words = set()
         self._frame_iterator = None
+        self.use_ram = False
+        self.label_id = 0
         self.logger = logging.getLogger('forth')
 
     def init(self):
         # load core language definitions from a forth file
         self._include('__init__.forth')
 
+
+    def look_up(self, word):
+        """Find the word in one of the name spaces for the host and return the value"""
+        # target words are included w/ least priority. they must be available
+        # so that compiling words on the host works
+        lowercased_word = word.lower() # case insensitive
+        for namespace in (self.namespace, self.builtins, self.target_namespace):
+            try:
+                element = namespace[lowercased_word]
+            except KeyError:
+                pass
+            else:
+                return element
+        raise KeyError('%r not in any namespace (host)' % (word,))
+
+    def look_up_target(self, word):
+        """Find the word in one of the namespaces for the target and return the value"""
+        # builtin namespace is not searched as it only includes words
+        # implemented in python. target name space has priority over normal
+        # space.
+        lowercased_word = word.lower() # case insensitive
+        for namespace in (self.target_namespace, self.namespace):
+            try:
+                element = namespace[lowercased_word]
+            except KeyError:
+                pass
+            else:
+                return element
+        raise KeyError('%r not in any namespace (target)' % (word,))
+
+    def create_label(self):
+        """Create a new assembler label"""
+        self.label_id += 1
+        return '__lbl%s' % (self.label_id,)
+
     def create_asm_label(self, name):
         """\
         There are a number of symbols that are not allowed in assembler
@@ -175,23 +369,23 @@ class Forth(rpn.RPN):
                 ('-', '_dash_'),
                 ('+', '_plus_'),
                 ('*', '_star_'),
+                ('?', '_qmark_'),
                 ('/', '_slash_'),
                 ('\\', '_backslash_'),
                 ('|', '_or_'),
                 ('&', '_and_'),
-                ('@', '_at_'),
+                ('@', '_fetch_'),
                 ('[', '_open_bracket_'),
                 (']', '_close_bracket_'),
-                ('!', '_excl_'),
+                ('!', '_store_'),
                 ('<', '_less_'),
                 ('>', '_bigger_'),
                 ('=', '_eq_'),
                 ('NOT', '_NOT_'),
-                #~ ('AND', '_AND_'),
-                #~ ('OR', '_OR_'),
+                ("'", '_tick_'),
         ):
             name = name.replace(t_in, t_out)
-        return name
+        return '_' + name
 
     def interpret_word(self, word):
         """Depending on mode a word is executed or compiled"""
@@ -248,7 +442,7 @@ class Forth(rpn.RPN):
         value = stack.pop()
         if isinstance(reference, Variable):
             if reference.frame != self.frame:
-                raise ValueError('!: Frame mismatch for variable %r != %r' % (reference.frmae, self.frame))
+                raise ValueError('!: Frame mismatch for variable %r != %r' % (reference.frame, self.frame))
             if isinstance(value, Variable):
                 reference.set(value.offset)
             else:
@@ -262,58 +456,6 @@ class Forth(rpn.RPN):
         stack.push(Variable(self.frame, len(self.frame)))
 
 
-    @rpn.word('/MOD')
-    def word_divmod(self, stack):
-        a = stack.pop()
-        b = stack.pop()
-        d, m = divmod(a, b)
-        stack.push(d)
-        stack.push(m)
-
-    @rpn.word('=')
-    def word_equals1(self, stack):
-        """Compare two numbers for equality"""
-        a = stack.pop()
-        b = stack.pop()
-        stack.push(a == b)
-
-    @rpn.word('0=')
-    def word_is_zero(self, stack):
-        """Check if number is not zero"""
-        a = stack.pop()
-        stack.push(a == 0)
-
-    @rpn.word('0>')
-    def word_is_positive(self, stack):
-        """Check if number is positive"""
-        a = stack.pop()
-        stack.push(a > 0)
-
-    @rpn.word('?DUP')
-    def word_Qdup(self, stack):
-        """DUP top of stack but only if not zero."""
-        if stack[-1]:
-            stack.push[-1]
-
-    @rpn.word('ROT')
-    def word_is_rot(self, stack):
-        """Rotate 3 items on the stack. 3rd gets 1st."""
-        a = stack.pop()
-        b = stack.pop()
-        c = stack.pop()
-        stack.push(b)
-        stack.push(a)
-        stack.push(c)
-
-    @rpn.word('-ROT')
-    def word_is_nrot(self, stack):
-        """Rotate 3 items on the stack. 1st gets 3rd."""
-        a = stack.pop()
-        b = stack.pop()
-        c = stack.pop()
-        stack.push(a)
-        stack.push(c)
-        stack.push(b)
 
     @immediate
     @rpn.word("'")
@@ -330,12 +472,20 @@ class Forth(rpn.RPN):
         """Push ASCII code of next character"""
         name = stack.next_word()
         value = ord(name[0])
+        stack.push(value)
+
+    @immediate
+    @rpn.word('[CHAR]')
+    def word_compile_char(self, stack):
+        """Compile ASCII code of next character"""
+        name = stack.next_word()
+        value = ord(name[0])
         if self.compiling:
             if self.frame is None: raise ValueError('not in colon definition')
             self.frame.append(self.instruction_literal)
             self.frame.append(value)
         else:
-            stack.push(value)
+            raise ValueError('interpretation sematics undefined')
 
     @rpn.word(',')
     def word_coma(self, stack):
@@ -351,7 +501,6 @@ class Forth(rpn.RPN):
     @rpn.word(':')
     def word_colon(self, stack):
         """Begin defining a function"""
-        if self.frame is not None: raise ValueError('already in colon definition')
         name = self.next_word()
         self.frame = Frame(name)
         self.compiling = True
@@ -371,7 +520,6 @@ class Forth(rpn.RPN):
     @rpn.word('CODE')
     def word_code(self, stack):
         """Begin defining a native code function"""
-        if self.frame is not None: raise ValueError('already in colon definition')
         name = self.next_word()
         self.frame = NativeFrame(name)
         self.compiling = True
@@ -382,26 +530,32 @@ class Forth(rpn.RPN):
         """End definition of a native code function"""
         if self.frame is None: raise ValueError('not in colon definition')
         #~ print "defined code", self.frame.name, self.frame     # XXX DEBUG
-        self.namespace[self.frame.name.lower()] = self.frame
-        self.frame = None
-        self.compiling = False
-
-    @immediate
-    @rpn.word('END-CODE-INTERNAL')
-    def word_end_code_internal(self, stack):
-        """\
-        End definition of a native code function. The name is stored in an
-        internal name space. This is used to implement words that are built-in /
-        native code on the host, which can not be cross compiled and which
-        should not be shadowed be definitions that only work for the target.
-        """
-        if self.frame is None: raise ValueError('not in colon definition')
-        #~ print "defined code internal", self.frame.name, self.frame     # XXX DEBUG
         self.target_namespace[self.frame.name.lower()] = self.frame
         self.frame = None
         self.compiling = False
 
 
+    @immediate
+    @rpn.word('INTERRUPT')
+    def word_interrupt(self, stack):
+        """Begin defining an interrupt function"""
+        name = self.next_word()
+        vector = self.pop()
+        self.frame = InterruptFrame(name, vector)
+        self.compiling = True
+
+    @immediate
+    @rpn.word('END-INTERRUPT')
+    def word_end_interrupt(self, stack):
+        """End definition of a native code function"""
+        if self.frame is None: raise ValueError('not in colon definition')
+        #~ print "defined code", self.frame.name, self.frame     # XXX DEBUG
+        self.target_namespace[self.frame.name.lower()] = self.frame
+        self.frame = None
+        self.compiling = False
+
+
+
     @immediate
     @rpn.word('IMMEDIATE')
     def word_immediate(self, stack):
@@ -417,6 +571,7 @@ class Forth(rpn.RPN):
         item = self.look_up(stack.next_word())
         self.frame.append(item)
 
+
     @immediate
     @rpn.word('[')
     def word_interpret(self, stack):
@@ -451,6 +606,7 @@ class Forth(rpn.RPN):
     @immediate
     @rpn.word('RECURSE')
     def word_recurse(self, stack):
+        """Call currently defined word"""
         if not self.compiling: raise ValueError('not allowed in immediate mode')
         if self.frame is None: raise ValueError('not in colon definition')
         # put conditional branch operation in sequence, remember position of offset on stack
@@ -461,54 +617,106 @@ class Forth(rpn.RPN):
 
     @rpn.word('WORD')
     def word_word(self, stack):
+        """Read next word from the source and put it on the stack."""
         stack.push(stack.next_word())
 
     @rpn.word('.')
     def word_dot(self, stack):
+        """Output element on stack"""
         self.output.write(unicode(stack.pop()))
 
     @rpn.word('EMIT')
     def word_emit(self, stack):
+        """Output number on stack as unicode character."""
         self.output.write(unichr(stack.pop()))
 
     @rpn.word('VARIABLE')
     def word_variable(self, stack):
-        """Allocate a variable. Creates space in RAM and a address getter function."""
+        """\
+        Allocate a variable. Creates space in RAM and an address getter
+        function.
+        """
         name = stack.next_word()
         # allocate separate memory for the variable
         # (cross compiled to RAM)
-        self.variables[name] = Frame()
+        self.variables[name] = Frame('var'+name)
         self.variables[name].append(0)
         # create a function that pushes the variables address
-        self.namespace[name] = frame
-        frame = Frame()
+        frame = Frame(name)
         frame.append(self.instruction_literal)
-        frame.append(self.look_up(name))
+        frame.append(self.variables[name])
+        self.namespace[name.lower()] = frame
+        # XXX could also do a native impl with "push #adr;NEXT"
 
     @rpn.word('VALUE')
     def word_value(self, stack):
-        """Allocate a variable. Creates space in RAM and a value getter function."""
+        """\
+        Allocate a variable. Creates space in RAM and a value getter
+        function.
+        """
+        value = stack.pop()
         name = stack.next_word()
         # allocate separate memory for the variable
         # (cross compiled to RAM)
-        self.variables[name] = Frame()
-        self.variables[name].append(0)
+        self.variables[name] = Frame('val'+name)
+        self.variables[name].append(value)
         # create a function that pushes the variables address
-        self.namespace[name] = frame
-        frame = Frame()
+        frame = Frame(name)
         frame.append(self.instruction_literal)
-        frame.append(self.look_up(name))
+        frame.append(self.variables[name])
         frame.append(self.look_up('@'))
+        self.namespace[name.lower()] = frame
 
+    @immediate
     @rpn.word('TO')
     def word_to(self, stack):
         """Write to a VALUE"""
         name = stack.next_word()
-        value = stack.pop()
-        self.variables[name][0] = value
+        if self.compiling:
+            self.frame.append(self.instruction_literal)
+            self.frame.append(self.variables[name])
+            self.frame.append(self.look_up('!'))
+        else:
+            value = stack.pop()
+            self.variables[name][0] = value # XXX
+
+    @rpn.word('RAM')
+    def word_ram(self, stack):
+        """Select RAM as target for following CREATE calls"""
+        self.use_ram = True
+
+    @rpn.word('ROM')
+    def word_rom(self, stack):
+        """Select ROM/Flash as target for following CREATE calls"""
+        self.use_ram = False
+
+    @rpn.word('CREATE')
+    def word_create(self, stack):
+        name = stack.next_word()
+        # allocate separate memory
+        # (cross compiled to RAM)
+        self.variables[name] = Frame('cre'+name)
+        self.variables[name].use_ram = self.use_ram
+        self.frame = self.variables[name]
+        # create a function that pushes the variables address
+        frame = Frame(name)
+        frame.append(self.instruction_literal)
+        frame.append(self.variables[name])
+        self.namespace[name.lower()] = frame
+        # XXX could also do a native impl with "push #adr;NEXT"
+
+    @rpn.word('ALLOT')
+    def word_allot(self, stack):
+        count = stack.pop()
+        if count > 0:
+            if count & 1: raise ValueError('odd sizes currently not supported')
+            self.frame.extend([0]*(count/2))
+        else:
+            raise ValueError('negative ALLOT not supported')
 
     @rpn.word('CONSTANT')
     def word_constant(self, stack):
+        """Declare a constant. Assign next word to value from stack."""
         value = stack.pop()
         name = stack.next_word()
         stack.namespace[name.lower()] = value
@@ -531,7 +739,11 @@ class Forth(rpn.RPN):
         words = []
         while True:
             word = self.next_word()
-            if word == '"': break
+            if word.endswith('"'):
+                # emulate character wise reading
+                if word != '"':
+                    words.append(word[:-1])
+                break
             words.append(word)
         text = codecs.escape_decode(u' '.join(words))[0]
         if self.compiling:
@@ -547,7 +759,11 @@ class Forth(rpn.RPN):
         words = []
         while True:
             word = self.next_word()
-            if word == '"': break
+            if word.endswith('"'):
+                # emulate character wise reading
+                if word != '"':
+                    words.append(word[:-1])
+                break
             words.append(word)
         text = codecs.escape_decode(u' '.join(words))[0]
         if self.compiling:
@@ -559,7 +775,10 @@ class Forth(rpn.RPN):
     @immediate
     @rpn.word('DEPENDS-ON')
     def word_depends_on(self, stack):
-        """Mark word as used so that it is included in cross compilation."""
+        """\
+        Mark word as used so that it is included in cross compilation. Useful
+        when using other words within CODE definitions.
+        """
         if self.compiling:
             word = self.next_word()
             self.frame.append(self.word_depends_on)
@@ -569,7 +788,10 @@ class Forth(rpn.RPN):
             self._compile_remember(word)
 
     def _compile_remember(self, word):
-        """Remember words used but yet compiled"""
+        """\
+        Remember that a word used. This ensures that it is included in the list
+        of cross compiled words.
+        """
         # track what is not yet done
         word = word.lower()
         if word not in self.compiled_words:
@@ -581,33 +803,60 @@ class Forth(rpn.RPN):
         remembered and can be output later, either manually with CROSS-COMPILE
         or automatically with CROSS-COMPILE-MISSING.
         """
+        self.output.write(u'.text\n.even\n')
         self.output.write(u';%s\n' % ('-'*76))
         self.output.write(u'; compilation of word %s\n' % frame.name)
         self.output.write(u';%s\n' % ('-'*76))
+        self.output.write(u'%s:\n' % self.create_asm_label(frame.name))
+        # compilation of the thread
+        self.output.write('\tbr #%s\n' % self.create_asm_label('DOCOL'))
+        #~ self.output.write('\tjmp %s\n' % self.create_asm_label('DOCOL'))
+        self._compile_thread(frame)
+        self.output.write('\t.word %s\n\n' % self.create_asm_label('EXIT'))
+
+    def _compile_thread(self, frame):
         next = iter(frame).next
         try:
-            self.output.write(u'%s:\n' % self.create_asm_label(frame.name))
-            self.output.write('\tjmp DOCOL\n')
             while True:
                 entry = next()
                 if callable(entry):
                     if entry == self.instruction_output_text:
+                        label = self.create_label()
+                        self.output.write('\t.word %s, %s\n' % (
+                                self.create_asm_label('__write_text'),
+                                self.create_asm_label(label)))
+                        self._compile_remember('__write_text')
+                        # output the text separately
+                        frame = NativeFrame(label)
+                        self.target_namespace[label] = frame
+                        self._compile_remember(label)
                         text = next()
-                        self.output.write(text)
-                        self.output.write('\n')
+                        frame.append(self.instruction_output_text)
+                        frame.append('\t.asciiz "%s"\n' % (codecs.escape_encode(text)[0],))
                     elif entry == self.instruction_literal:
                         value = next()
-                        self.output.write('\t.word LIT, %s\n' % value)
+                        if isinstance(value, Frame):
+                            self.output.write('\t.word %s, %s\n' % (
+                                    self.create_asm_label('LIT'),
+                                    self.create_asm_label(value.name),))
+                        else:
+                            self.output.write('\t.word %s, %-6s ; 0x%04x\n' % (
+                                    self.create_asm_label('LIT'),
+                                    value,
+                                    value))
                         self._compile_remember('LIT')
                     elif entry == self.instruction_seek:
+                        # branch needs special case as offset needs to be recalculated
                         offset = next()
-                        self.output.write('\t.word BRANCH, %s\n' % (offset*2,))
+                        self.output.write('\t.word %s, %s\n' % (self.create_asm_label('BRANCH'), offset*2))
                         self._compile_remember('BRANCH')
                     elif entry == self.instruction_branch_if_false:
+                        # branch needs special case as offset needs to be recalculated
                         offset = next()
-                        self.output.write('\t.word BRANCH0, %s\n' % (offset*2,))
+                        self.output.write('\t.word %s, %s\n' % (self.create_asm_label('BRANCH0'), offset*2))
                         self._compile_remember('BRANCH0')
                     elif hasattr(entry, 'rpn_name'):
+                        # for built-ins just take the name of the function
                         self.output.write('\t.word %s\n' % self.create_asm_label(entry.rpn_name.upper()))
                         self._compile_remember(entry.rpn_name)
                     elif isinstance(entry, (Frame, NativeFrame)):
@@ -616,23 +865,50 @@ class Forth(rpn.RPN):
                     else:
                         raise ValueError('Cross compilation undefined for %r' % entry)
                 else:
-                    raise ValueError('Cross compilation undefined for %r' % entry)
+                    self.output.write('\t.word %r\n' % (entry,))
+                    #~ raise ValueError('Cross compilation undefined for %r' % entry)
         except StopIteration:
             pass
-        self.output.write('\t.word EXIT\n\n')
 
     def _compile_native_frame(self, frame):
         """Compilation of native code function"""
+        self.output.write(u'.text\n.even\n')
         self.output.write(u';%s\n' % ('-'*76))
         self.output.write(u'; compilation of native word %s\n' % frame.name)
         self.output.write(u';%s\n' % ('-'*76))
         self.output.write(u'%s:\n' % self.create_asm_label(frame.name))
-        # native code blocks are executed to get the output
+        # native code blocks are executed. They are expected to print out
+        # assembler code
         frame(self)
-        self.output.write('\n')
+        self.output.write('\n') # get some space between this and next word
+
+    def _compile_interrupt_frame(self, frame):
+        """Compilation of interrupt function"""
+        self.output.write(u'.text\n.even\n')
+        self.output.write(u';%s\n' % ('-'*76))
+        self.output.write(u'; compilation of interrupt %s\n' % frame.name)
+        self.output.write(u';%s\n' % ('-'*76))
+
+        # interrupt entry code
+        self.output.write(u'vector_%s:\n' % (frame.vector))
+        self.output.write(u'\tsub #4, RTOS     ; prepare to push 2 values on return stack\n')
+        self.output.write(u'\tmov IP, 2(RTOS)  ; save IP on return stack\n')
+        self.output.write(u'\tmov SP, 0(RTOS)  ; save SP pointer on return stack it points to SR on stack\n')
+        self.output.write(u'\tmov #%s, IP      ; Move address of thread of interrupt handler in IP\n' % self.create_asm_label(frame.name))
+        self.output.write('\tjmp %s\n' % self.create_asm_label('DO-INTERRUPT'))
+        # the thread for the interrupt handler
+        self.output.write(u'%s:\n' % self.create_asm_label(frame.name))
+        self._compile_thread(frame)
+        self.output.write('\t.word %s\n\n' % self.create_asm_label('EXIT-INTERRUPT'))
+        self._compile_remember('DO-INTERRUPT')
+        self._compile_remember('EXIT-INTERRUPT')
 
 
     def instruction_cross_compile(self, stack, word=None):
+        """\
+        Cross compile word. This function can be called directly (w/ word
+        parameter) or be part of a Frame.
+        """
         if word is None:
             word = self._frame_iterator.next()
         # when interpreting, execute the actual functionality
@@ -641,12 +917,14 @@ class Forth(rpn.RPN):
         if word in self.not_yet_compiled_words:
             self.not_yet_compiled_words.remove(word)
         # get the frame and compile it - prefer target_namespace
-        if word in self.target_namespace:
-            item = self.target_namespace[word]
-        else:
-            item = self.look_up(word)
+        try:
+            item = self.look_up_target(word)
+        except KeyError:
+            raise ValueError('word %r is not available on the target' % (word,))
         # translate, depending on type
-        if isinstance(item, Frame):
+        if isinstance(item, InterruptFrame):
+            self._compile_interrupt_frame(item)
+        elif isinstance(item, Frame):
             self._compile_frame(item)
         elif isinstance(item, NativeFrame):
             self._compile_native_frame(item)
@@ -659,10 +937,11 @@ class Forth(rpn.RPN):
         """Output cross compiled version of function."""
         word = self.next_word()
         if self.compiling:
-            # when compiling add call to self and the word
+            # when compiling add call to self and the word to the Frame
             self.frame.append(self.instruction_cross_compile)
             self.frame.append(word)
         else:
+            # in interpretation mode, compile it now
             self.instruction_cross_compile(stack, word)
 
     @rpn.word('CROSS-COMPILE-MISSING')
@@ -678,19 +957,18 @@ class Forth(rpn.RPN):
     @rpn.word('CROSS-COMPILE-VARIABLES')
     def word_cross_compile_variables(self, stack):
         """\
-        Compile all the words that are used by other compiled words but not
-        yet translated. While compiling words, new words can be found which are
-        then also compiled.
+        Output section with variables (values in RAM).
         """
         self.output.write(u';%s\n' % ('-'*76))
         self.output.write(u'; Variables\n')
         self.output.write(u';%s\n' % ('-'*76))
         self.output.write(u'.bss\n')
-        for variable in self.variables:
+        # XXX check .use_ram attribute
+        for name, variable in sorted(self.variables.items()):
             variable.name
             self.output.write(u'%s:  .skip %d \n' % (
                     self.create_asm_label(variable.name),
-                    len(variable)))
+                    2*len(variable)))
             self.output.write('\n')
         self.output.write(u'.text\n')
 
@@ -702,6 +980,7 @@ class Forth(rpn.RPN):
         self._include(name)
 
     def _include(self, name):
+        """Include given filename. The Forth code is directly executed."""
         if name not in self.included_files:
             for prefix in self.include_path:
                 path = os.path.join(prefix, name)
@@ -725,7 +1004,7 @@ class Forth(rpn.RPN):
 
     @rpn.word('SHOW')
     def word_SHOW(self, stack):
-        """Show internals of given word"""
+        """Show internals of given word. Used to debug."""
         name = self.next_word()
         sys.stderr.write('SHOW %r\n' % name)
         try:
