@@ -17,16 +17,11 @@ from msp430.asm import mcu_definition_parser
 from msp430.asm import rpn, peripherals
 
 
-class LinkError(Exception):
+class LinkError(rpn.RPNError):
     """\
     Exception class for errors that can occur during linking. The exception can
     be annotated with the location in the source file.
     """
-    def __init__(self, message, filename=None, lineno=None, column=0):
-        Exception.__init__(self, message)
-        self.filename = filename
-        self.lineno = lineno
-        self.column = column
 
 
 # a segment has a valid range of addresses that is set by
@@ -202,6 +197,8 @@ class Linker(rpn.RPN):
         # the linking will require multiple passes, a flag controls
         # if errors are fatal or ignored
         self.errors_are_fatal = True
+        # to check labels for duplicate definition
+        self.check_labels = None
         # The link instructions
         self.instructions = instructions
         # information about the input
@@ -329,14 +326,26 @@ class Linker(rpn.RPN):
         if name in self.labels:
             if self.labels[name] != value:
                 raise LinkError('redefinition of symbol with different value (previous: %r, new: %r)' % (
-                    self.labels[name],
-                    value))
+                            self.labels[name],
+                            value),
+                        self.source_filename,
+                        self.source_line,
+                        self.source_column)
         self.labels[name] = value
 
     @rpn.word('CREATE-SYMBOL')
     def _create_symbol(self, rpn):
         """Mark current location with symbol"""
         name = self.name_symbol(self.next_word())
+        #~ # this simple check does not work as we're doing multiple passes
+        if self.check_labels is not None:
+            if name in self.check_labels:
+                raise LinkError(
+                        'Label %r redefined (old value: %r)' % (name, self.labels[name]),
+                        self.source_filename,
+                        self.source_line,
+                        self.source_column)
+            self.check_labels[name] = self.address
         self.labels[name] = self.address
 
     @rpn.word('GET-SYMBOL')
@@ -483,7 +492,9 @@ class Linker(rpn.RPN):
         This run is used to find all labels at their final locations.
         """
         self.top_segment.clear()
+        self.check_labels = {}
         self.interpret_sequence(self.instructions)
+        self.check_labels = None
         # create automatic labels for all segments (start/end)
         for segment in self.segments.values():
             name = segment.name.replace('.', '')    # remove dots in names
@@ -557,9 +568,12 @@ def to_TI_Text(segments):
 
 # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 
-if __name__ == '__main__':
+def main():
     import os
+    import logging
     from optparse import OptionParser
+    logging.basicConfig()
+
     parser = OptionParser(usage="""\
 %prog [options] [FILE...]|-]
 
@@ -595,10 +609,10 @@ Output is in "TI-Text" format."""
 
     parser.add_option(
             "-v", "--verbose",
-            action="store_true",
+            action="count",
             dest="verbose",
-            default=False,
-            help="print status messages")
+            default=0,
+            help="print status messages, gan be given multiple times to increase messages")
 
     parser.add_option(
             "--debug",
@@ -614,6 +628,13 @@ Output is in "TI-Text" format."""
 
     (options, args) = parser.parse_args()
 
+    if options.debug:
+        logging.getLogger().setLevel(logging.DEBUG)
+    elif options.verbose:
+        logging.getLogger().setLevel(logging.INFO)
+    else:
+        logging.getLogger().setLevel(logging.WARN)
+
     # prepare output
     if options.outfile is not None:
         out = open(options.outfile, "wb")
@@ -626,22 +647,26 @@ Output is in "TI-Text" format."""
     instructions = []
     for filename in args:
         if filename == '-':
-            if options.verbose:
+            if options.verbose > 2:
                 sys.stderr.write(u'reading stdin...\n')
             instructions.append('reset')
             instructions.extend(sys.stdin.read().split())
         else:
-            if options.verbose:
+            if options.verbose > 2:
                 sys.stderr.write(u'reading file "%s"...\n'% filename)
             instructions.append('reset')
-            instructions.extend(rpn.words_in_file(filename))
+            try:
+                instructions.extend(rpn.words_in_file(filename))
+            except IOError, e:
+                sys.stderr.write('ld: %s: File not found\n' % (filename,))
+                sys.exit(1)
 
     linker = Linker(instructions)
 
     # load symbols
     if options.symbols is not None:
-        peripherals = peripherals.load_internal(options.symbols)
-        for peripheral in peripherals.peripherals.values():
+        all_peripherals = peripherals.load_internal(options.symbols)
+        for peripheral in all_peripherals.peripherals.values():
             for reg_name, register in peripheral.items():
                 if reg_name.startswith('__'): continue
                 if '__address__' in register:
@@ -656,7 +681,7 @@ Output is in "TI-Text" format."""
 
     # ========= load MCU definition =========
 
-    if options.verbose:
+    if options.verbose > 1:
         sys.stderr.write("Step 1: load segment descriptions.\n")
 
     # load the file and get the desired MCU description
@@ -665,6 +690,7 @@ Output is in "TI-Text" format."""
             mem_maps = mcu_definition_parser.load_from_file(options.segmentfile)
         else:
             mem_maps = mcu_definition_parser.load_internal()
+        options.mcu_name = options.mcu_name.upper() # XXX hack
         segment_definitions = mcu_definition_parser.expand_definition(mem_maps, options.mcu_name)
     except Exception, msg:
         sys.stderr.write('ERROR loading segment descriptions: %s\n' % (msg,))
@@ -672,31 +698,31 @@ Output is in "TI-Text" format."""
 
     linker.segments_from_definition(segment_definitions)
 
-    if options.verbose:
+    if options.verbose > 2:
         sys.stderr.write('Segments available:\n')
         linker.top_segment.print_tree(sys.stderr)
 
     # ========= Do the actual linking =========
 
     try:
-        if options.verbose:
+        if options.verbose > 1:
             sys.stderr.write("Step 2: generate machine code\n")
             sys.stderr.write("        Pass 1: determinate segment sizes.\n")
         linker.pass_one()
 
-        if options.verbose:
+        if options.verbose > 1:
             sys.stderr.write("        Pass 2: calculate labels.\n")
         linker.pass_two()
 
-        if options.verbose:
+        if options.verbose > 1:
             sys.stderr.write("        Pass 3: final output.\n")
         linker.pass_three()
-    except rpn.RPNError, e:
+    except rpn.RPNError as e:
         sys.stderr.write(u"%s:%s: %s\n" % (e.filename, e.lineno, e))
         if options.debug and e.text:
             sys.stderr.write(u"%s:%s: input line was: %r\n" % (e.filename, e.lineno, e.text))
         sys.exit(1)
-    except LinkError, msg:
+    except LinkError as msg:
         if msg.filename and msg.lineno is not None:
             sys.stderr.write(u'%s:%s: %s\n' % (msg.filename, msg.lineno, msg))
         else:
@@ -705,12 +731,12 @@ Output is in "TI-Text" format."""
 
     # ========= Output final result =========
 
-    if options.verbose:
+    if options.verbose > 1:
         sys.stderr.write("Step 3: write machine code to file.\n")
 
     out.write(to_TI_Text(linker.segments))
 
-    if options.verbose:
+    if options.verbose > 1:
         sys.stderr.write("Labels:\n")
         labels = linker.labels.keys()
         labels.sort()
@@ -730,3 +756,6 @@ Output is in "TI-Text" format."""
         linker.top_segment.sort_subsegments(by_address=True)
         linker.top_segment.print_tree(sys.stderr, hide_empty=True)
 
+
+if __name__ == '__main__':
+    main()
